from dataclasses import Field, dataclass
from enum import Enum
from typing import Any, ClassVar, Generic, Protocol, TypedDict, TypeVar, cast

from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCollection,
)
from pymongo.client_session import ClientSession
from pymongo.collection import Collection

from . import exceptions


class Dataclass(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


DTO = TypeVar('DTO', bound=Dataclass)
CollectionType = TypeVar('CollectionType', AsyncIOMotorCollection, Collection[Any])
SessionType = TypeVar('SessionType', AsyncIOMotorClientSession, ClientSession)


@dataclass(repr=False, slots=True, eq=False)
class Index:
    """Allows to specify extra information about mongodb index."""
    field: str
    name: str | None = None
    desc: bool = True
    unique: bool = False


class Access(int, Enum):
    """Use to indicate method access in a repository."""
    PUBLIC = 0
    PROTECTED = 1
    PRIVATE = 2


class MetaAttributes(Generic[CollectionType], TypedDict, total=True):
    dto: type[Dataclass] | None
    collection: CollectionType | None
    index: str | Index | None
    method_access: Access | None
    id_field: str | None


class CollectionProvider(Generic[CollectionType]):
    def __init__(self, obj: Any, collection: CollectionType | None = None):
        self.collection: CollectionType | None = collection
        self.obj = obj

    def provide(self) -> CollectionType:
        # First check if collection already provided
        if self.collection is not None:
            return self.collection

        # Check if collection present in object attributes
        if (__mongorepo__ := getattr(self.obj, '__mongorepo__', None)) is not None:
            collection: CollectionType = __mongorepo__['collection_provider'].collection
            if collection is not None:
                self.collection = collection
                return collection

        # Check if collection in Meta class
        meta = self.obj.__dict__.get('Meta', None)
        if meta is not None:
            if (c := getattr(meta, 'collection', None)) is not None:
                if not isinstance(c, (AsyncIOMotorCollection, Collection)):
                    raise exceptions.NoCollectionException
                self.collection = cast(CollectionType, c)
                return self.collection
        # If no collection raise exception
        raise exceptions.NoCollectionException('Collection cannot be found', with_meta=False)


class CollectionProviderDescriptor(Generic[CollectionType]):
    """Descriptor to get mongo collection."""

    def __init__(self, collection: CollectionType | None = None):
        self.collection: CollectionType | None = collection

    def __get__(self, obj: object, owner: type) -> CollectionType:
        # First check if collection already provided
        if self.collection is not None:
            return self.collection

        # Check if collection present in object attributes
        if (__mongorepo__ := getattr(obj, '__mongorepo__', None)) is not None:
            collection: CollectionType = __mongorepo__['collection_provider'].collection
            if collection is not None:
                self.collection = collection
                return collection

        # Check if collection in Meta class
        meta = obj.__dict__.get('Meta', None)
        if meta is not None:
            if (c := getattr(meta, 'collection', None)) is not None:
                if not isinstance(c, (AsyncIOMotorCollection, Collection)):
                    raise exceptions.NoCollectionException
                self.collection = cast(CollectionType, c)
                return self.collection
        # If no collection raise exception
        raise exceptions.NoCollectionException('Collection cannot be found', with_meta=False)



from typing import Callable

from ._handlers import _handle_implement_specific_methods
from .methods import SpecificMethod


def implement[T: type](*specific_methods: SpecificMethod) -> Callable[[T], T]:
    """Decorator that allows to implement methods of `base_cls`

    ## Usage example:

    ```
    class IRepo:
        async def get(self, title: str) -> SomeDataclass:
            ...

        async def add(self, model: SomeDataclass) -> SomeDataclass:
            ...

    @implement(
        GetMethod(IRepo.get, filters=['title']),
        AddMethod(IRepo.add, dto='model'),
    )
    class MongoRepo:
        class Meta:
            collection = some_collection()
            dto = SomeDataclass

    r: IRepo = MongoRepo()
    await r.add(SomeDataclass('some title'))
    dto = await r.get(title='some title')
    print(dto)  # SomeDataclass(title='some title')
    ```

    """
    if not specific_methods:
        raise ValueError('No methods to implement')

    def wrapper(cls: T) -> T:
        return _handle_implement_specific_methods(cls, *specific_methods)
    return wrapper



import inspect
from typing import Any, Callable, Protocol

from mongorepo.modifiers.base import ModifierAfter, ModifierBefore

from ._types import LParameter, MethodAction, ParameterEnum

Modifiers = list[ModifierBefore | ModifierAfter]


class FieldAlias:
    """Class that allow to set alias for `dataclass` field
    ### Usage example:
    ```
    @dataclass
    class User:
        name: str

    #                 User.name = username
    username_alias = FieldAlias('name', 'username')

    class UserRepository(ABC):
        @abstractmethod
        def get_user(self, username: str) -> User | None:
            ...

    @implement(GetMethod(UserRepository.get_user, filters=[username_alias]))
    class MongoUserRepository:
        ...

    repo: UserRepository = MongoUserRepository()
    user = repo.get_user(username='admin')
    print(user)  # User(name='admin')
    ```
    """

    __slots__ = ('name', 'aliases')

    def __init__(self, field: str, *aliases: str) -> None:
        self.name = field
        self.aliases = aliases

    def is_alias(self, string: str) -> bool:
        return string in self.aliases

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        if isinstance(other, FieldAlias):
            return self.name == other.name and self.aliases == other.aliases
        return False

    def __repr__(self) -> str:
        aliases = ', '.join([f'"{a}"' for a in self.aliases])
        return f'{self.__class__.__name__}("{self.name}", {aliases})'

    def __str__(self) -> str:
        return self.name


class _ManageMethodFiltersMixin:
    @staticmethod
    def manage_filters(filters: list[str | FieldAlias]) -> dict[str | ParameterEnum, Any]:
        params = {f: ParameterEnum.FILTER for f in filters if isinstance(f, str)}
        aliases = {
            ParameterEnum.FILTER_ALIAS: dict.fromkeys(a.aliases, a.name)
            for a in filters if isinstance(a, FieldAlias)
        }
        return {**params, **aliases}  # type: ignore


class SpecificMethod(Protocol):
    """Protocol for specific methods.

    ### Implementations::

        | import from               | class
        mongorepo.implement.methods.GetMethod
        mongorepo.implement.methods.GetListMethod
        mongorepo.implement.methods.GetAllMethod
        mongorepo.implement.methods.AddMethod
        mongorepo.implement.methods.AddBatchMethod
        mongorepo.implement.methods.UpdateMethod
        mongorepo.implement.methods.DeleteMethod

        mongorepo.implement.methods.IncrementIntegerFieldMethod

        mongorepo.implement.methods.ListAppendMethod
        mongorepo.implement.methods.ListPopMethod
        mongorepo.implement.methods.ListRemoveMethod
        mongorepo.implement.methods.ListGetFieldValuesMethod

    """
    name: str
    source: Callable
    params: dict[str, LParameter]
    action: MethodAction
    modifiers: Modifiers

    @property
    def is_async(self) -> bool:
        ...


class SpecificFieldMethod(SpecificMethod):
    field_name: str
    integer_weight: int | None


class Method:
    """Base class that represents mongorepo method."""

    def __init__(
        self,
        source: Callable,
        **params: LParameter,
    ) -> None:
        self.source: Callable = source
        self.params: dict[str, LParameter] = params
        self.name: str = source.__name__

    def __repr__(self) -> str:
        params_repr = ', '.join(f'{k}={v}' for k, v in self.params.items())
        return (
            f'Method({self.source}, {self.name}, {params_repr})'
        )

    @property
    def signature(self) -> inspect.Signature:
        return inspect.signature(self.source)

    def get_source_params(self, exclude_self: bool = True) -> dict:
        gen_params = dict(self.signature.parameters)
        if exclude_self:
            gen_params.pop('self')
        return gen_params

    @property
    def is_async(self) -> bool:
        return inspect.iscoroutinefunction(self.source)


class GetMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `get` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    @dataclass
    class User:
        id: str

    class Repo(typing.Protocol):
        # this method can be also asynchronous
        def get(self, id: str) -> User:
            ...

    @implement(GetMethod(Repo.get, filters=['id']))
    class MongoRepo:
        class Meta:
            dto = User
            collection = mongo_client().users.users_collection

    repo = MongoRepo()
    user = repo.get(id='123')
    ```

    """

    def __init__(
        self,
        source: Callable,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(source, **self.manage_filters(filters))  # type: ignore
        self.action = MethodAction.GET
        self.modifiers = modifiers or []


class AddMethod(Method):
    """
    ### Class that represents mongorepo `add` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    @dataclass
    class User:
        name: str

    class Repo(typing.Protocol):
        # this method can be also asynchronous
        def add(self, user: User) -> User:
            ...

    @implement(AddMethod(Repo.add, dto='user'))
    class MongoRepo:
        class Meta:
            dto = User
            collection = mongo_client().users.users_collection

    repo = MongoRepo()
    added_user = repo.add(user=User(name='admin'))
    ```

    """

    def __init__(
        self,
        source: Callable,
        dto: str,
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(source, **{dto: 'dto'})  # type: ignore
        self.action = MethodAction.ADD
        self.modifiers = modifiers or []


class UpdateMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `update` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    @dataclass
    class User:
        id: str
        name: str

    @dataclass
    class UpdateUser:
        name: str

    class Repo(typing.Protocol):
        # this method can be also asynchronous
        def update_user(self, id: str, update_model: UpdateUser) -> User:
            ...

    @implement(UpdateMethod(Repo.update_user, filters=['id'], dto='update_model'))
    class MongoRepo:
        class Meta:
            dto = User
            collection = mongo_client().users.users_collection

    repo = MongoRepo()
    updated_user: User = repo.update_user(id='123', UpdateUser(name='admin_1'))
    print(updated_user.name)  # admin_1
    ```
    ## Note
    * All fields of update model will be used to update record in DB

    ### Example:

    ```
    user = repo.get(id='1')
    print(user)  # User(name='admin', id='1')

    @dataclass
    class UpdateUser:
        id: str | None = None
        name: str | None = None

    # In this example the desired action is to update only `name` field, but
    # in practice, together with `name` will be updated `id`
    updated_user = repo.update(id='1', UpdateUser(name='creator'))
    print(updated_user)  # User(id=None, name='creator')

    # To avoid it use UpdateSkipModifier class
    @implement(
        UpdateMethod(modifiers=[UpdateSkipModifier(skip_if_value=None)], ...)
    )
    class MongoRepo: ...

    updated_user = repo.update(id='1', UpdateUser(name='creator'))
    print(updated_user)  # User(id='1', name='creator')
    ```

    """
    def __init__(
        self,
        source: Callable,
        dto: str,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(
            source, **{dto: 'dto'}, **self.manage_filters(filters),  # type: ignore
        )
        self.action = MethodAction.UPDATE
        self.modifiers = modifiers or []


class DeleteMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `delete` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    class Repo(typing.Protocol):
        # this method can be also asynchronous
        def remove_user(self, id: str) -> bool:
            ...

    @implement(DeleteMethod(Repo.remove_user, filters=['id']))
    class MongoRepo:
        class Meta:
            dto = User
            collection = mongo_client().users.users_collection

    repo = MongoRepo()
    deleted: bool = repo.remove_user(id='1')
    ```

    """
    def __init__(
        self,
        source: Callable,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(source, **self.manage_filters(filters))
        self.action = MethodAction.DELETE
        self.modifiers = modifiers or []


class GetListMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `get_list` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    class BookRepo(typing.Protocol):
        # this method can be also asynchronous
        def get_list_of_books(self, category: str) -> list[Book]:
            ...

    @implement(GetListMethod(BookRepo.get_list_of_books, filters=['category']))
    class MongoRepo:
        class Meta:
            dto = Book
            collection = mongo_client().books.books_collection

    repo = MongoRepo()
    books = repo.get_list_of_books(category='fiction')
    print(books)  # [Book(title='...', category='fiction'), Book(title='...', category='fiction')]
    ```

    """
    def __init__(
        self,
        source: Callable,
        filters: list[FieldAlias | str],
        offset: str,
        limit: str,
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(
            source, **{offset: 'offset', limit: 'limit'},  # type: ignore
            **self.manage_filters(filters),
        )
        self.action = MethodAction.GET_LIST
        self.modifiers = modifiers or []


class GetAllMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `get_all` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    class BookRepo(typing.Protocol):
        # this method can be also asynchronous
        def get_all_books(self) -> typing.Generator[Book, None]:
            ...

    @implement(GetAllMethod(BookRepo.get_all, filters=[]))
    class MongoRepo:
        class Meta:
            dto = Book
            collection = mongo_client().books.books_collection

    repo = MongoRepo()
    books = repo.get_all_books()
    print(next(books))  # Book(title='...', category='fiction')
    ```
    """

    def __init__(
        self,
        source: Callable,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(source, **self.manage_filters(filters))
        self.action = MethodAction.GET_ALL
        self.modifiers = modifiers or []


class AddBatchMethod(Method):
    """
    ### Class that represents mongorepo `add_batch` method
    * supports `async`, `await` syntax
    ## Usage example

    ```
    class BookRepo(typing.Protocol):
        # this method can be also asynchronous
        def add_books(self, books: list[Book]) -> None:
            ...

    @implement(AddBatchMethod(BookRepo.add_books, dto_list=['books']))
    class MongoRepo:
        class Meta:
            dto = Book
            collection = mongo_client().books.books_collection

    repo = MongoRepo()
    repo.add_books([
        Book(title='1', category='science fiction'),
        Book(title='2', category='romance'),
    ])
    ```
    """

    def __init__(
        self,
        source: Callable,
        dto_list: str,
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(source, **{dto_list: 'dto_list'})  # type: ignore
        self.action = MethodAction.ADD_BATCH
        self.modifiers = modifiers or []


class ListAppendMethod(Method, _ManageMethodFiltersMixin):
    """### Class that represents mongorepo `list_append` method

    * supports `async`, `await` syntax
    * works this dataclasses and standard types (e.g. str, int, etc.)
    ## Usage example

    ```

    @dataclass
    class Box:
        weight: int

    @dataclas
    class Cargo:
        id: str
        boxes: list[Box]

    class CargoRepo(typing.Protocol):
        # this method can be also asynchronous
        def add_box_to_cargo(self, id: str, box: Box) -> None:
            ...

    @implement(
        ListAppendMethod(
            CargoRepo.add_box_to_cargo,
            field_name='boxes',  # Cargo.boxes
            filters=['id'],
            value='box'
        )
    )
    class MongoRepo:
        class Meta:
            dto = Cargo
            collection = mongo_client().cargo.cargo_collection

    repo = MongoRepo()
    repo.add_box_to_cargo(id='1', box=Box(weight=5))
    cargo = repo.get(id='1')
    print(cargo)  # Cargo(id='1', boxes=[Box(weight=5)])
    ```

    """

    def __init__(
        self,
        source: Callable,
        field_name: str,
        value: str,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(
            source, **{value: 'value'}, **self.manage_filters(filters),  # type: ignore
        )
        self.field_name = field_name
        self.action = MethodAction.LIST_APPEND
        self.modifiers = modifiers or []


class ListPopMethod(Method, _ManageMethodFiltersMixin):
    """### Class that represents mongorepo `list_pop` method

    * supports `async`, `await` syntax
    * works this dataclasses and standard types (e.g. str, int, etc.)
    ## Usage example

    ```

    @dataclass
    class Box:
        weight: int

    @dataclas
    class Cargo:
        id: str
        boxes: list[Box]

    class CargoRepo(typing.Protocol):
        # this method can be also asynchronous
        def pop_box(self, id: str) -> Box:  # or raises mongorepo.exceptions.NotFoundException
            ...

    @implement(
        ListPopMethod(
            CargoRepo.pop_box,
            field_name='boxes',  # Cargo.boxes
            filters=['id'],
        ),
        ...
    )
    class MongoRepo:
        class Meta:
            dto = Cargo
            collection = mongo_client().cargo.cargo_collection

    repo = MongoRepo()
    cargo = repo.get(id='1')
    print(cargo)  # Cargo(id='1', boxes=[Box(weight=5)])
    box = repo.pop_box(id='1')
    print(box)  # Box(weight=5)
    print(repo.get(id='id'))  # Cargo(id='1', boxes=[])
    ```

    """

    def __init__(
        self,
        source: Callable,
        field_name: str,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(
            source, **self.manage_filters(filters),
        )
        self.field_name = field_name
        self.action = MethodAction.LIST_POP
        self.modifiers = modifiers or []


class ListRemoveMethod(Method, _ManageMethodFiltersMixin):
    """### Class that represents mongorepo `list_remove` method

    * supports `async`, `await` syntax
    * works this dataclasses and standard types (e.g. str, int, etc.)
    ## Usage example

    ```

    @dataclass
    class Box:
        weight: int

    @dataclas
    class Cargo:
        id: str
        boxes: list[Box]

    class CargoRepo(typing.Protocol):
        # this method can be also asynchronous
        def remove_box(
            self, id: str, box: Box,
        ) -> Box:  # or raises mongorepo.exceptions.NotFoundException
            ...

    @implement(
        ListRemoveMethod(
            CargoRepo.pop_box,
            field_name='boxes',  # Cargo.boxes
            value='box'
            filters=['id'],
        ),
        ...
    )
    class MongoRepo:
        class Meta:
            dto = Cargo
            collection = mongo_client().cargo.cargo_collection

    repo = MongoRepo()
    cargo = repo.get(id='1')
    print(cargo)  # Cargo(id='1', boxes=[Box(weight=5)])
    repo.remove_box(id='1', box=Box(weight=5))
    print(repo.get(id='id'))  # Cargo(id='1', boxes=[])
    ```

    """

    def __init__(
        self,
        source: Callable,
        field_name: str,
        value: str,
        filters: list[FieldAlias | str],
        modifiers: Modifiers | None = None,
    ) -> None:
        super().__init__(
            source, **{value: 'value'}, **self.manage_filters(filters),  # type: ignore
        )
        self.field_name = field_name
        self.action = MethodAction.LIST_REMOVE
        self.modifiers = modifiers or []


class ListGetFieldValuesMethod(Method, _ManageMethodFiltersMixin):
    """### Class that represents mongorepo `list_field_values` method

    * supports `async`, `await` syntax
    * works this dataclasses and standard types (e.g. str, int, etc.)
    ## Usage example

    ```

    @dataclass
    class Box:
        weight: int

    @dataclas
    class Cargo:
        id: str
        boxes: list[Box]

    class CargoRepo(typing.Protocol):
        # this method can be also asynchronous
        def get_cargo_boxes(
            self, id: str, limit: int = 5,
        ) -> list[Box]:  # or raises mongorepo.exceptions.NotFoundException
            ...

    @implement(
        ListGetFieldValuesMethod(
            CargoRepo.get_cargo_boxes,
            field_name='boxes',  # Cargo.boxes
            filters=['id'],
            limit='limit',
        ),
        ...
    )
    class MongoRepo:
        class Meta:
            dto = Cargo
            collection = mongo_client().cargo.cargo_collection

    repo = MongoRepo()
    cargo = repo.get(id='1')
    print(cargo)  # Cargo(id='1', boxes=[ Box(weight=5), Box(weight=2), Box(weight=1) ])
    boxes = repo.get_cargo_boxes(id='1', limit=2))
    print(boxes)  # [ Box(weight=5), Box(weight=2) ]
    ```

    """

    def __init__(
        self,
        source: Callable,
        field_name: str,
        filters: list[FieldAlias | str],
        offset: str | None = None,
        limit: str | None = None,
        modifiers: Modifiers | None = None,
    ) -> None:
        params: dict[str, Any] = {}
        if offset:
            params[offset] = 'offset'
        if limit:
            params[limit] = 'limit'
        super().__init__(
            source,
            **params,
            **self.manage_filters(filters),
        )
        self.field_name = field_name
        self.action = MethodAction.LIST_FIELD_VALUES
        self.modifiers = modifiers or []


class IncrementIntegerFieldMethod(Method, _ManageMethodFiltersMixin):
    """
    ### Class that represents mongorepo `integer_increment` method
    * supports `async`, `await` syntax
    ## Usage example

    ```

    @dataclass
    class Record:
        id: str
        views: int

    class RecordRepo(typing.Protocol):
        # this method can be also asynchronous
        def increment_views(
            self, id: str,
        ) -> None:  # or raises mongorepo.exceptions.NotFoundException
            ...

    @implement(
        IncrementIntegerFieldMethod(
            RecordRepo.increment_views,
            field_name='views',  # Record.views
            filters=['id'],
        ),
        ...
    )
    class MongoRepo:
        class Meta:
            dto = Record
            collection = mongo_client().records.record_collection

    repo = MongoRepo()
    repo.add(Record(id='1', views=1000))
    repo.increment_views(id='1')
    print(repo.get(id='1'))  # Record(id='1', views=1001)
    ```
    """

    def __init__(
        self,
        source: Callable,
        field_name: str,
        filters: list[FieldAlias | str],
        weight: str | None = None,
        default_weight_value: int = 1,
        modifiers: Modifiers | None = None,
    ) -> None:
        params = {} if weight is None else {weight: 'weight'}
        super().__init__(
            source, **params, **self.manage_filters(filters),  # type: ignore
        )
        self.action = MethodAction.INTEGER_INCREMENT
        self.field_name = field_name
        self.integer_weight = default_weight_value
        self.modifiers = modifiers or []



from contextlib import contextmanager
from typing import Any, Generic, Protocol, TypedDict

from ._base import CollectionProvider, CollectionType, SessionType
from ._methods.interfaces import MongorepoMethod


class MongorepoDict(Generic[SessionType, CollectionType], TypedDict, total=True):
    methods: dict[str, MongorepoMethod[SessionType]]
    collection_provider: CollectionProvider[CollectionType]


class HasMongorepoDict(Generic[SessionType, CollectionType], Protocol):
    __mongorepo__: MongorepoDict[SessionType, CollectionType]


def default_mongorepo_dict(
    collection_provider: CollectionProvider[CollectionType],
    methods: dict[str, MongorepoMethod[SessionType]] | None = None,
) -> MongorepoDict[SessionType, CollectionType]:
    methods = methods or {}
    return MongorepoDict(collection_provider=collection_provider, methods=methods)


def use_session(
    session: SessionType,
    *mongorepo_repositories: HasMongorepoDict[SessionType, CollectionType] | Any,
):
    """Function to make mongorepo methods use session."""
    for repo in mongorepo_repositories:
        __mongorepo__: MongorepoDict[SessionType, CollectionType] | None = getattr(
            repo, '__mongorepo__', None,
        )
        if not __mongorepo__:
            raise TypeError(
                f'Invalid class for mongorepo repository: {type(repo)}: '
                f'"{type(repo).__name__}" does not implement {str(HasMongorepoDict)} protocol',
            )

        for method in __mongorepo__['methods'].values():
            method.session = session


def remove_session(
    *mongorepo_repositories: HasMongorepoDict[SessionType, CollectionType] | Any,
):
    """Function to remove session from mongorepo methods."""
    for repo in mongorepo_repositories:
        __mongorepo__: MongorepoDict[SessionType, CollectionType] | None = getattr(
            repo, '__mongorepo__', None,
        )
        if not __mongorepo__:
            raise TypeError(
                f'Invalid class for mongorepo repository: {type(repo)}: '
                f'"{type(repo).__name__}" does not implement {str(HasMongorepoDict)} protocol',
            )

        for method in __mongorepo__['methods'].values():
            method.session = None


@contextmanager
def session_context(
    session: SessionType,
    *mongorepo_repositories: HasMongorepoDict[SessionType, CollectionType] | Any,
):
    """Context manager that temporarily assigns a session to mongorepo methods.

    This ensures that all methods in the provided mongorepo repositories
    use the given session during the execution of the context. Once the
    context exits, the session is removed from the methods.

    Usage example::

        with session_context(session, repo1, repo2):
            repo1.some_method()  # Uses the provided session
            repo2.another_method()  # Uses the provided session

    Args:
        session (SessionType): The session to be assigned to mongorepo methods.
        *mongorepo_repositories (HasMongorepoDict[SessionType, CollectionType] | Any):
            One or more repository instances that implement the `HasMongorepoDict`
            protocol, allowing them to have session-managed methods.

    Yields:
        None: The context executes with the assigned session, then removes it upon exit.

    """
    try:
        use_session(session, *mongorepo_repositories)
        yield
    finally:
        remove_session(*mongorepo_repositories)



import asyncio
from typing import Any, AsyncGenerator, Generator, Generic

from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCollection,
)
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.results import InsertManyResult

from mongorepo import exceptions
from mongorepo._base import DTO, Dataclass, Index, MetaAttributes
from mongorepo._methods.impl import (
    AddBatchMethod,
    AddMethod,
    DeleteMethod,
    GetAllMethod,
    GetListMethod,
    GetMethod,
    UpdateMethod,
)
from mongorepo._methods.impl_async import (
    AddBatchMethodAsync,
    AddMethodAsync,
    DeleteMethodAsync,
    GetAllMethodAsync,
    GetListMethodAsync,
    GetMethodAsync,
    UpdateMethodAsync,
)
from mongorepo.utils import (
    CollectionProvider,
    _create_index,
    _create_index_async,
    _get_converter,
    _get_dto_from_origin,
    _get_meta_attributes,
)

from ._common import MongorepoDict


class BaseMongoRepository(Generic[DTO]):
    """
    ## Base MongoDB repository class
    #### Extends child classes with various methods:

    ```
    add(self, dto: DTO) -> DTO
    get(self, **filters) -> DTO | None
    get_all(self, **filters) -> Iterable[DTO]
    update(self, dto: DTO, **filters) -> DTO
    delete(self, **filters) -> bool
    ```

    #### Provide DTO type in type hints, example:

    ```
    class DummyMongoRepository(BaseMongoRepository[UserDTO]):
        ...
    ```

    * If you want to create an index use `mongorepo.Index`
      or just a name of the field to put index on
    """

    def __new__(cls, *args, **kwargs) -> 'BaseMongoRepository[DTO]':
        instance = super().__new__(cls)
        try:
            dto_type = _get_dto_from_origin(cls)
        except exceptions.NoDTOTypeException:
            dto_type = None

        try:
            meta: MetaAttributes[Collection] | None = _get_meta_attributes(cls)
        except exceptions.NoMetaException:
            meta = None

        id_field: str | None = meta['id_field'] if meta else None

        if dto_type is None:
            dto_type = meta['dto'] if meta else None
            if not dto_type:
                raise exceptions.NoDTOTypeException

        collection: Collection | None = meta['collection'] if meta else None
        index: Index | str | None = meta['index'] if meta else None

        if index is not None:
            if collection is None:
                raise exceptions.NoCollectionException(
                    message='Index can be created only if collection provided in Meta class',
                )
            _create_index(index, collection=collection)
        converter = _get_converter(dto_type, id_field)

        add_method = AddMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        add_batch_method = AddBatchMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_method = GetMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_list_method = GetListMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_all_method = GetAllMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        update_method = UpdateMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        delete_method = DeleteMethod(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa

        if hasattr(cls, '__mongorepo__'):
            __mongorepo__: MongorepoDict[ClientSession, Collection] = getattr(cls, '__mongorepo__')
            __mongorepo__['methods']['add'] = add_method
            __mongorepo__['methods']['add_batch'] = add_batch_method
            __mongorepo__['methods']['get'] = get_method
            __mongorepo__['methods']['get_list'] = get_list_method
            __mongorepo__['methods']['get_all'] = get_all_method
            __mongorepo__['methods']['update'] = update_method
            __mongorepo__['methods']['delete'] = delete_method
        else:
            __mongorepo__ = MongorepoDict[ClientSession, Collection](
                collection_provider=CollectionProvider(obj=cls, collection=collection),
                methods={
                    'add': add_method,
                    'add_batch': add_batch_method,
                    'get': get_method,
                    'get_list': get_list_method,
                    'get_all': get_all_method,
                    'update': update_method,
                    'delete': delete_method,
                },
            )
            cls.__mongorepo__ = __mongorepo__

        setattr(cls, '_mongorepo_add', __mongorepo__['methods']['add'])
        setattr(cls, '_mongorepo_add_batch', __mongorepo__['methods']['add_batch'])
        setattr(cls, '_mongorepo_get', __mongorepo__['methods']['get'])
        setattr(cls, '_mongorepo_get_list', __mongorepo__['methods']['get_list'])
        setattr(cls, '_mongorepo_get_all', __mongorepo__['methods']['get_all'])
        setattr(cls, '_mongorepo_update', __mongorepo__['methods']['update'])
        setattr(cls, '_mongorepo_delete', __mongorepo__['methods']['delete'])

        return instance

    def get(self, **filters: Any) -> DTO | None:
        return self._mongorepo_get(**filters)  # type: ignore

    def get_all(self, **filters: Any) -> Generator[DTO, None, None]:
        yield from self._mongorepo_get_all(**filters)  # type: ignore

    def get_list(self, offset: int = 0, limit: int = 20, **filters: Any) -> list[DTO]:
        return self._mongorepo_get_list(offset=offset, limit=limit, **filters)  # type: ignore

    def add(self, dto: DTO) -> DTO:
        return self._mongorepo_add(dto)  # type: ignore

    def add_batch(self, dto_list: list[DTO]) -> 'InsertManyResult':
        return self._mongorepo_add_batch(dto_list)  # type: ignore

    def update(self, dto: Dataclass, **filters: Any) -> DTO | None:
        return self._mongorepo_update(dto, **filters)  # type: ignore

    def delete(self, **filters: Any) -> bool:
        return self._mongorepo_delete(**filters)  # type: ignore


class BaseAsyncMongoRepository(Generic[DTO]):
    """
    ## Base MongoDB repository class
    #### Extends child classes with various methods:

    ```
    add(self, dto: DTO) -> DTO
    get(self, **filters) -> DTO | None
    get_all(self, **filters) -> Iterable[DTO]
    update(self, dto: DTO, **filters) -> DTO
    delete(self, **filters) -> bool
    ```

    #### Provide DTO type in type hints, example:

    ```
    class DummyMongoRepository(BaseMongoRepository[UserDTO]):
        ...
    ```

    * If you want to create an index use `mongorepo.Index`
      or just a name of the field to put index on
    """

    def __new__(cls, *args, **kwargs) -> 'BaseAsyncMongoRepository[DTO]':
        instance = super().__new__(cls)
        try:
            dto_type = _get_dto_from_origin(cls)
        except exceptions.NoDTOTypeException:
            dto_type = None

        try:
            meta: MetaAttributes[AsyncIOMotorCollection] | None = _get_meta_attributes(cls)
        except exceptions.NoMetaException:
            meta = None

        id_field: str | None = meta['id_field'] if meta else None

        if dto_type is None:
            dto_type = meta['dto'] if meta else None
            if not dto_type:
                raise exceptions.NoDTOTypeException

        collection: AsyncIOMotorCollection | None = meta['collection'] if meta else None
        index: Index | str | None = meta['index'] if meta else None

        if index is not None:
            if collection is None:
                raise exceptions.NoCollectionException(
                    message='Index can be created only if collection provided in Meta class',
                )
            asyncio.create_task(_create_index_async(index, collection=collection))
        converter = _get_converter(dto_type, id_field)

        add_method = AddMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        add_batch_method = AddBatchMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_method = GetMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_list_method = GetListMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        get_all_method = GetAllMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        update_method = UpdateMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa
        delete_method = DeleteMethodAsync(dto_type, cls, id_field=id_field, converter=converter)  # type: ignore  # noqa

        if hasattr(cls, '__mongorepo__'):
            __mongorepo__: MongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection] = getattr(cls, '__mongorepo__')  # noqa
            __mongorepo__['methods']['add'] = add_method
            __mongorepo__['methods']['add_batch'] = add_batch_method
            __mongorepo__['methods']['get'] = get_method
            __mongorepo__['methods']['get_list'] = get_list_method
            __mongorepo__['methods']['get_all'] = get_all_method
            __mongorepo__['methods']['update'] = update_method
            __mongorepo__['methods']['delete'] = delete_method
        else:
            __mongorepo__ = MongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection](
                collection_provider=CollectionProvider(obj=cls, collection=collection),
                methods={
                    'add': add_method,
                    'add_batch': add_batch_method,
                    'get': get_method,
                    'get_list': get_list_method,
                    'get_all': get_all_method,
                    'update': update_method,
                    'delete': delete_method,
                },
            )
            cls.__mongorepo__ = __mongorepo__

        setattr(cls, '_mongorepo_add', __mongorepo__['methods']['add'])
        setattr(cls, '_mongorepo_add_batch', __mongorepo__['methods']['add_batch'])
        setattr(cls, '_mongorepo_get', __mongorepo__['methods']['get'])
        setattr(cls, '_mongorepo_get_list', __mongorepo__['methods']['get_list'])
        setattr(cls, '_mongorepo_get_all', __mongorepo__['methods']['get_all'])
        setattr(cls, '_mongorepo_update', __mongorepo__['methods']['update'])
        setattr(cls, '_mongorepo_delete', __mongorepo__['methods']['delete'])

        return instance

    async def get(self, **filters: Any) -> DTO | None:
        return await self.__class__._mongorepo_get(**filters)  # type: ignore

    def get_all(self, **filters: Any) -> AsyncGenerator[DTO, None]:
        return self._mongorepo_get_all(**filters)  # type: ignore

    async def get_list(self, offset: int = 0, limit: int = 20, **filters: Any) -> list[DTO]:
        return await self._mongorepo_get_list(offset=offset, limit=limit, **filters)  # type: ignore

    async def add(self, dto: DTO) -> DTO:
        return await self._mongorepo_add(dto)  # type: ignore

    async def add_batch(self, dto_list: list[DTO]) -> 'InsertManyResult':
        return await self._mongorepo_add_batch(dto_list)  # type: ignore

    async def update(self, dto: Dataclass, **filters: Any) -> DTO | None:
        return await self._mongorepo_update(dto, **filters)  # type: ignore

    async def delete(self, **filters: Any) -> bool:
        return await self._mongorepo_delete(**filters)  # type: ignore



import inspect
import types
import warnings
from dataclasses import is_dataclass
from functools import partial
from typing import (
    Any,
    Callable,
    NoReturn,
    get_args,
    get_origin,
    get_type_hints,
)

import pymongo
from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import (
    DTO,
    Access,
    CollectionProvider,
    Dataclass,
    Index,
    MetaAttributes,
)


def raise_exc(exc: Exception | type[Exception]) -> NoReturn:
    """Allows to write one-lined exceptions."""
    raise exc


def get_prefix(access: Access | None, cls: type | None = None) -> str:
    """
    Returns string prefix according to Access value,
    * it can be `'_'`, `'__'`, `_{cls.__name__}__` or `''`
    """
    match access:
        case Access.PRIVATE:
            prefix = f'_{cls.__name__}__' if cls else '__'
        case Access.PROTECTED:
            prefix = '_'
        case Access.PUBLIC | None:
            prefix = ''
    return prefix


def _get_meta(cls: type) -> type:
    """Tries to get `Meta` class for the class or raises `NoMetaException`
    exception."""
    meta = cls.__dict__.get('Meta', None)
    if not meta:
        raise exceptions.NoMetaException
    if not isinstance(meta, type):
        raise exceptions.NoMetaException
    return meta


def _get_meta_attributes(cls: type) -> MetaAttributes:
    """Collect all available attributes from `Meta` class."""
    meta = _get_meta(cls)

    dto_type: type[Dataclass] | None = getattr(meta, 'dto', None)

    collection: AsyncIOMotorCollection | Collection[Any] | None = getattr(meta, 'collection', None)

    index: Index | str | None = getattr(meta, 'index', None)

    method_access: Access | None = getattr(meta, 'method_access', None)

    substitute: dict[str, str] | None = getattr(meta, 'substitute', None)
    if substitute is not None:
        warnings.warn("'substitute' dictionary is deprecated, please pass arguments in other place")

    id_field: str | None = getattr(meta, 'id_field', None)

    return MetaAttributes(
        dto=dto_type,
        collection=collection,  # type: ignore
        index=index,
        method_access=method_access,
        substitute=substitute,
        id_field=id_field,
    )


def get_first_arg(type_hint) -> Any:
    first_list_arg = get_args(type_hint)[0]
    if type(first_list_arg) is types.UnionType:
        first_list_arg = get_args(first_list_arg)[0]
    elif type(first_list_arg) is types.GenericAlias:
        first_list_arg = get_origin(first_list_arg)
    return first_list_arg


def get_dataclass_type_hints(dataclass: type[Dataclass]) -> dict[str, Any]:
    type_hints: dict[str, Any] = {}

    for field_name, hint in get_type_hints(dataclass).items():
        if is_dataclass(hint):
            type_hints[field_name] = hint
        elif (org := get_origin(hint)) is list:
            first_left_arg = get_first_arg(hint)
            type_hints[field_name] = first_left_arg
        elif org is types.UnionType:
            type_hints[field_name] = get_args(hint)[0]
        else:
            type_hints[field_name] = hint

    return type_hints


def _create_index(index: Index | str, collection: Collection) -> None:
    """### Creates an index for the collection

    * index parameter can be string or `mongorepo.Index`
    * If index is string, creates standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings

    """
    if isinstance(index, str):
        collection.create_index(index)
        return
    index_name = index.name or f'index_{index.field}'
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )


async def _create_index_async(index: Index | str, collection: AsyncIOMotorCollection) -> None:
    """### Creates an index for the collection

    * index parameter can be string or mongorepo.Index
    * If index is string, create standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings

    """
    if isinstance(index, str):
        await collection.create_index(index)
        return
    index_name = f'index_{index.field}'
    if index.name:
        index_name = index.name
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    await collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )


def _get_dto_from_origin(cls: type) -> type[Dataclass]:
    """
    Tries to get `dto` from origin of the class or raises an exception

    ```
    class A[UserDTO]: ...

    _get_dto_from_origin(A)  # UserDTO

    ```
    """
    try:
        if not hasattr(cls, '__orig_bases__'):
            raise exceptions.NoDTOTypeException
        dto: type = get_args(cls.__orig_bases__[0])[0]
    except IndexError:
        raise exceptions.NoDTOTypeException
    if dto is DTO:
        raise exceptions.NoDTOTypeException
    if not is_dataclass(dto):
        raise exceptions.NoDTOTypeException

    return dto


def _convert_to_dto(dto_type: type[DTO], dct: dict[str, Any]) -> DTO:
    """Converts document to dto, does not include mongodb `_id`"""
    dct.pop('_id') if dct.get('_id', None) else ...
    return dto_type(**dct)


def _convert_to_dto_with_id(
    id_field: str,
) -> Callable:
    """Converts document to dto, includes mongodb `_id` allows to set specific
    field where to store `_id`"""
    def wrapper(dto_type: type[DTO], dct: dict[str, Any]) -> DTO:
        dct[id_field] = str(dct.pop('_id'))
        return dto_type(**dct)
    return wrapper


def _nested_convert_to_dto[T: Dataclass](
    dataclass_type: type[T], data: dict[str, Any], id_field: str | None = None,
) -> T:
    def convert(
        data: dict[str, Any], dataclass_type: type[T], as_dataclass: bool = True,
    ) -> dict[str, Any] | T:
        type_hints = get_dataclass_type_hints(dataclass_type)
        result = {}
        for key, value in data.items():
            is_dtcls = is_dataclass(h := type_hints[key])

            if is_dtcls and isinstance(value, list):
                result[key] = [convert(v, h, True) for v in value]
            elif is_dtcls:
                result[key] = convert(value, h, True)  # type: ignore
            else:
                result[key] = value

        return dataclass_type(**result) if as_dataclass else result

    result = {}
    if id_field is not None:
        result[id_field] = str(data.pop('_id'))
    else:
        data.pop('_id') if data.get('_id', None) else ...

    result.update(convert(data, dataclass_type, False))  # type: ignore[arg-type]
    return dataclass_type(**result)


def _get_converter(
    dto_type: type[DTO], id_field: str | None = None,
) -> Callable[[type[DTO], dict[str, Any]], DTO]:
    """Returns proper dataclass converter based on type hints of the dataclas.

    ## Usage example::

        from mongorepo import get_converter

        @dataclass
        class User:
            id: str
            username: str
            friends: list['User'] = field(default_factory=list)

        to_user = functools.partial(get_converter(User), User)  # just not to pass User every time

        dct = {
            'id': '1',
            'username': 'admin',
            'friends': [
                {'id': 2, 'username': 'bob', 'friends': []},
                {'id': 3, 'username': 'destroyer', 'friends': [
                        {'id': 4, 'username': 'top_1', 'friends': []}
                    ]
                }
            ]
        }
        user = to_user(dct)

        pprint.pprint(user)
        # User(id='1',
        #     username='admin',
        #     friends=[User(id=2, username='bob', friends=[]),
        #             User(id=3,
        #                 username='destroyer',
        #                 friends=[User(id=4, username='top_1', friends=[])])])

    """

    converter: Callable[[type[DTO], dict[str, Any]], DTO] | partial = _convert_to_dto
    r = _has_dataclass_fields(dto_type=dto_type)
    if r:
        converter = partial(_nested_convert_to_dto, id_field=id_field)
    elif id_field is not None:
        converter = _convert_to_dto_with_id(id_field=id_field)
    return converter


def _has_dataclass_fields(dto_type: type[DTO]) -> bool:
    type_hints = get_dataclass_type_hints(dto_type)
    for v in type_hints.values():
        if is_dataclass(v):
            return True
    return False


def _check_valid_field_type(field_name: str, dto_type: type[DTO], data_type: type) -> None:
    field = dto_type.__annotations__.get(field_name, None)
    if field is None:
        raise exceptions.MongorepoException(
            message=f'{dto_type} does not have field "{field_name}"',
        )
    org = get_origin(field)
    if field == data_type or org is data_type:
        return
    elif type(org) is types.UnionType:
        union_args = get_args(org)
        if union_args[0] != data_type:
            raise exceptions.MongorepoException(
                message=f'Invalid type of the field "{field_name}", expected: {data_type}',
            )
    raise exceptions.MongorepoException(
        message=f'Invalid type of the field "{field_name}", expected: {data_type}',
    )


def _get_defaults(func: Callable) -> dict[str, Any]:
    """Return default values for function parameters if any provided."""
    result = {}
    params = inspect.signature(func).parameters
    for param in params.values():
        if param.default is not inspect._empty:
            result[param.name] = param.default
    return result


def use_collection[T](
    collection: AsyncIOMotorCollection | Collection[Any],
) -> Callable[[T], T]:
    """
    Simple decorator to make mongorepo use specific collection,
    useful for dynamic look up of collection

    ```
    # 1. Valid
    # @use_collection(my_collection)
    @mongorepo.repository(add=True, get=True)
    # 2. Valid
    # @use_collection(my_collection)
    class Repository:
        class Meta:
            dto = SimpleDTO

    def provide_repository() -> Repository:
        # 3. Valid
        return use_collection(my_collection)(Repository)()

    repo = provide_repository()
    ```
    """
    def wrapper(cls: T) -> T:
        provider = CollectionProvider(cls, collection)  # type: ignore
        if (__mongorepo__ := getattr(cls, '__mongorepo__', None)) is not None:
            __mongorepo__['collection_provider'] = provider  # type: ignore
        else:
            setattr(cls, '__mongorepo__', {'collection_provider': provider, 'methods': {}})
        return cls
    return wrapper



import typing as t
from dataclasses import asdict, is_dataclass

from bson import ObjectId
from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCollection,
)
from pymongo.results import InsertManyResult, UpdateResult

from mongorepo._base import Dataclass
from mongorepo._common import HasMongorepoDict
from mongorepo.modifiers.base import ModifierAfter, ModifierBefore
from mongorepo.utils import _get_converter, get_dataclass_type_hints


class AddMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.id_field = id_field
        self.kwargs = kwargs

    async def __call__(self, dto: T) -> T:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            dto = modifier_before.modify(dto=dto)

        extra = {}
        object_id = ObjectId()
        if self.id_field:
            dto.__dict__[self.id_field] = str(object_id)
            extra['_id'] = object_id
        await collection.insert_one({**asdict(dto), **extra}, session=self.session)

        for modifier_after in self.modifiers_after:
            dto = modifier_after.modify(dto)

        return dto


class AddBatchMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.id_field = id_field
        self.kwargs = kwargs

    async def __call__(self, dto_list: list[T]) -> InsertManyResult:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            dto_list = modifier_before.modify(dto_list=dto_list)

        if self.id_field:
            batch: list[dict[str, t.Any]] = []
            for dto in dto_list:
                object_id = ObjectId()
                dto.__dict__[self.id_field] = str(object_id)
                batch.append({**asdict(dto), '_id': object_id})
            result = await collection.insert_many(batch)
        else:
            result = await collection.insert_many(asdict(d) for d in dto_list)

        for modifier_after in self.modifiers_after:
            result = modifier_after.modify(result)

        return result


class GetAllMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.kwargs = kwargs

    async def __call__(self, **filters: t.Any) -> t.AsyncGenerator[T, None]:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            filters = modifier_before.modify(**filters)

        cursor = collection.find(filters)
        async for dct in cursor:
            yield self.converter(self.dto_type, dct)


class GetListMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.kwargs = kwargs

    async def __call__(self, offset: int = 0, limit: int = 20, **filters: t.Any) -> list[T]:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            offset, limit, filters = modifier_before.modify(
                offset, limit, **filters,
            )

        cursor = collection.find(filter=filters).skip(offset).limit(limit)
        result = [self.converter(self.dto_type, doc) async for doc in cursor]

        for modifier_after in self.modifiers_after:
            result = modifier_after.modify(result)

        return result


class GetMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.id_field = id_field
        self.kwargs = kwargs

    async def __call__(self, **filters: t.Any) -> T | None:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            filters = modifier_before.modify(**filters)

        result = await collection.find_one(filters)
        dto = self.converter(self.dto_type, result) if result else None

        for modifier_after in self.modifiers_after:
            dto = modifier_after.modify(dto)

        return dto


class DeleteMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.kwargs = kwargs

    async def __call__(self, **filters: t.Any) -> bool:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            filters = modifier_before.modify(**filters)

        deleted = await collection.find_one_and_delete(filters, session=self.session)

        for modifier_after in self.modifiers_after:
            deleted = modifier_after.modify(deleted)

        return True if deleted else False


class UpdateMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        converter: t.Callable[[type[T], dict[str, t.Any]], T] | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.owner = owner
        self.session: AsyncIOMotorClientSession | None = None
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.converter = converter or _get_converter(dto_type, id_field)
        self.kwargs = kwargs

    async def __call__(self, dto: T, **filters: t.Any) -> T | None:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            dto, filters = modifier_before.modify(dto=dto, **filters)

        data: dict[str, dict[str, t.Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            data['$set'][field] = value
        updated_document: dict[str, t.Any] | None = await collection.find_one_and_update(
            filter=filters, update=data, return_document=True,
        )

        result = self.converter(self.dto_type, updated_document) if updated_document else None

        for modifier_after in self.modifiers_after:
            result = modifier_after.modify(result)

        return result


class UpdateListFieldMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        action: t.Literal['$push', '$pull'],
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.field_name = field_name
        fields = get_dataclass_type_hints(dto_type)
        self.field_type = fields.get(field_name, None)
        self.field_converter = None
        if is_dataclass(self.field_type):
            self.field_converter = _get_converter(fields[field_name])
        self.owner = owner
        self.action = action
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.id_field = id_field
        self.kwargs = kwargs

    async def __call__(self, value: t.Any, **filters: t.Any) -> UpdateResult:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            value, filters = modifier_before.modify(value, **filters)

        value = value if not is_dataclass(self.field_type) else asdict(value)

        res = await collection.update_one(
            filter=filters,
            update={self.action: {self.field_name: value}},
            session=self.session,
        )

        for modifier_aftert in self.modifiers_after:
            res = modifier_aftert.modify(res)

        return res


class AppendListMethodAsync[T: Dataclass](UpdateListFieldMethodAsync[T]):
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            dto_type=dto_type,
            owner=owner,
            field_name=field_name,
            action='$push',
            modifiers=modifiers,
            session=session,
            id_field=id_field,
            **kwargs,
        )

    async def __call__(self, value: t.Any, **filters: t.Any) -> UpdateResult:
        return await super().__call__(value, **filters)


class RemoveListMethodAsync[T: Dataclass](UpdateListFieldMethodAsync[T]):
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            dto_type=dto_type,
            owner=owner,
            field_name=field_name,
            action='$pull',
            modifiers=modifiers,
            session=session,
            id_field=id_field,
            **kwargs,
        )

    async def __call__(self, value: t.Any, **filters: t.Any) -> UpdateResult:
        return await super().__call__(value, **filters)


class GetListValuesMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.field_name = field_name
        fields = get_dataclass_type_hints(dto_type)
        self.field_type = fields.get(field_name, None)
        self.owner = owner
        self.field_converter = None
        if is_dataclass(self.field_type):
            self.field_converter = _get_converter(fields[field_name])
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.kwargs = kwargs

    async def __call__(
        self, offset: int, limit: int, **filters: t.Any,
    ) -> list[T] | list[t.Any] | None:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            offset, limit, filters = modifier_before.modify(
                offset, limit, **filters,
            )

        document = await collection.find_one(
            filters, {self.field_name: {'$slice': [offset, limit]}},
        )
        if document is None:
            result = None

        elif is_dataclass(self.field_type):
            result = [
                self.field_converter(self.field_type, d)  # type: ignore
                for d in document[self.field_name]
            ]
        else:
            result = document[self.field_name]

        for modifier_aftert in self.modifiers_after:
            result = modifier_aftert.modify(result)

        return result


class PopListMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        **kwargs,
    ) -> None:
        self.dto_type = dto_type
        self.field_name = field_name
        self.owner = owner
        fields = get_dataclass_type_hints(dto_type)
        self.field_type = fields.get(field_name, None)
        self.field_converter = None
        if is_dataclass(self.field_type):
            self.field_converter = _get_converter(fields[field_name])
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.kwargs = kwargs

    async def __call__(self, **filters: t.Any) -> T | t.Any:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            filters = modifier_before.modify(**filters)

        document = await collection.find_one_and_update(
            filter=filters, update={'$pop': {self.field_name: 1}},
        )
        if document is None:
            result = None
        elif is_dataclass(self.field_type):
            result = self.field_converter(  # type: ignore
                self.field_type, document[self.field_name][-1],  # type: ignore[arg-type]
            )
        else:
            result = document[self.field_name][-1]

        for modifier_aftert in self.modifiers_after:
            result = modifier_aftert.modify(result)

        return result


class IncrementIntegerFieldMethodAsync[T: Dataclass]:
    def __init__(
        self,
        dto_type: type[T],
        owner: HasMongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection],
        field_name: str,
        weight: int = 1,
        modifiers: tuple[ModifierBefore | ModifierAfter, ...] = (),
        session: AsyncIOMotorClientSession | None = None,
        id_field: str | None = None,
        **kwargs,
    ) -> None:
        self.field_name = field_name
        self.owner = owner
        self.weight = weight
        self.session = session
        self.modifiers_after = [m for m in modifiers if isinstance(m, ModifierAfter)]
        self.modifiers_before = [m for m in modifiers if isinstance(m, ModifierBefore)]
        self.kwargs = kwargs

    async def __call__(self, weight: int | None = None, **filters) -> UpdateResult:
        collection = self.owner.__mongorepo__['collection_provider'].provide()

        for modifier_before in self.modifiers_before:
            weight, filters = modifier_before.modify(weight, **filters)

        w = weight if weight is not None else self.weight
        result = await collection.update_one(
            filter=filters, update={'$inc': {self.field_name: w}},
        )

        for modifier_aftert in self.modifiers_after:
            result = modifier_aftert.modify(result)

        return result



from . import decorators, exceptions
from ._base import DTO, Access, Index
from ._common import remove_session, session_context, use_session
from .classes import BaseAsyncMongoRepository, BaseMongoRepository
from .decorators import async_mongo_repository as async_repository
from .decorators import mongo_repository as repository
from .queries import AggregationStage, Condition, Operation, UpdateModifier
from .utils import _get_converter as get_converter
from .utils import use_collection

__all__ = [
    'get_converter',
    'use_collection',
    'use_session',
    'remove_session',
    'session_context',
    'UpdateModifier',
    'AggregationStage',
    'Operation',
    'Condition',
    'BaseMongoRepository',
    'BaseAsyncMongoRepository',
    'repository',
    'async_repository',
    'Access',
    'DTO',
    'Index',
    'decorators',
    'exceptions',
]



import inspect
from typing import Any, Callable

from mongorepo import exceptions
from mongorepo._base import Dataclass
from mongorepo.utils import _get_defaults

from ._types import MethodAction
from ._types import ParameterEnum as MongorepoParameter
from ._utils import implement_mapper, initialize_callable_mongorepo_method
from .methods import Method, SpecificFieldMethod, SpecificMethod


def _substitute_specific_method(
    cls,
    method: SpecificMethod | SpecificFieldMethod,
    dto_type: type[Dataclass],
    id_field: str | None = None,
    field_name: str | None = None,
    integer_weight: int | None = None,
    **kwargs,
) -> Callable:
    is_async = inspect.iscoroutinefunction(method.source)

    if hasattr(method, 'field_name'):
        field_name = method.field_name  # type: ignore

    if hasattr(method, 'integer_weight'):
        integer_weight = method.integer_weight  # type: ignore

    mapped_method = implement_mapper(method)
    callable_mongorepo_method = initialize_callable_mongorepo_method(
        mongorepo_method=mapped_method,
        dto_type=dto_type,
        owner=cls,
        id_field=id_field,
        field_name=field_name,
        integer_weight=integer_weight,
        modifiers=method.modifiers,
    )

    def func(self, *args, **kwargs) -> Any:
        required_params = _manage_custom_params(
            Method(source=method.source, **method.params), *args, **kwargs,
        )
        return callable_mongorepo_method(**required_params)

    async def async_func(self, *args, **kwargs) -> Any:
        required_params = _manage_custom_params(
            Method(source=method.source, **method.params), *args, **kwargs,
        )
        return await callable_mongorepo_method(**required_params)

    if method.action == MethodAction.GET_ALL and is_async is True:
        new_method = func
    else:
        new_method = async_func if is_async else func

    new_method.__annotations__ = method.source.__annotations__
    new_method.__name__ = method.name

    return new_method


def _manage_custom_params(
    method: Method,
    *args,
    **kwargs,
) -> dict[str, Any]:
    defaults = _get_defaults(method.source)
    result: dict[str, Any] = {}
    filters: dict[str, Any] = {}
    source_params_map: dict[str, Any] = {}
    aliases: dict[str, Any] = method.params.get(MongorepoParameter.FILTER_ALIAS, {})  # type: ignore
    i = 0

    # Iterate over all names of the method's arguments
    for source_param in method.get_source_params().keys():

        # Parameter was passed as positional argument
        if len(args) > i:
            source_params_map[source_param] = args[i]
            i += 1
            continue

        # Parameter was passed as keyword argument ?
        elif source_param in kwargs:
            source_params_map[source_param] = kwargs[source_param]

        # If parameter was not passed check for defaults
        elif defaults.get(source_param, None):
            source_params_map[source_param] = defaults[source_param]

        # Missing parameter
        else:
            raise exceptions.MongorepoException(
                message=f'Cannot find value for {source_param} parameter, '
                f'{method.name}() parameters: {method.params}',
            )

    for key, value in source_params_map.items():
        if method.params.get(key, None) == MongorepoParameter.FILTER:
            filters[key] = value
        elif (dto_field := aliases.get(key, None)) is not None:
            filters[dto_field] = value
        else:
            result[method.params[key]] = value

    result.update(filters)
    return result



class MongorepoException(Exception):
    def __init__(self, message: str | None = None):
        self.message = message

    def __str__(self) -> str:
        return self.message or super().__str__()


class NoMetaException(MongorepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Class does not have "Meta" class inside'


class NoDTOTypeException(MongorepoException):
    def __init__(self, message: str | None = None, with_meta: bool = True):
        self.with_meta: bool = with_meta
        self.message = message

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'DTO type was not provided' + ' in "Meta" class' if self.with_meta else ''


class NoCollectionException(MongorepoException):
    def __init__(self, message: str | None = None, with_meta: bool = True):
        self.with_meta: bool = with_meta
        self.message = message

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Collection was not provided' + ' in "Meta" class' if self.with_meta else ''


class NotDataClass(MongorepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Provided dto type does not implement dataclass interface'


class InvalidMethodNameException(MongorepoException):
    def __init__(
        self,
        method_name: str,
        message: str | None = None,
        available_methods: tuple[str, ...] | None = None,
    ):
        self.message = message
        self.method_name: str = method_name
        self.available_methods: tuple[str, ...] | None = available_methods

    def __str__(self) -> str:
        if self.message:
            return self.message
        message = f'Method "{self.method_name}" does not appear in mongorepo available methods'
        if self.available_methods:
            message += f'\nAvailable methods: {self.available_methods!r}'
        return message


class TypeHintException(MongorepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        message = 'Invalid type hint'
        return message


class NotFoundException(MongorepoException):
    def __init__(self, **filters) -> None:
        self.filters = filters

    def __str__(self) -> str:
        filters = ', '.join([f'{key}={value}' for key, value in self.filters.items()])
        return f'Document not found, filters: {filters}'


class InvalidActionException(MongorepoException):
    def __init__(self, valid_actions: list[str], action: str, message: str | None = None):
        self.message = message
        self.action: str = action
        self.valid_actions: list[str] = valid_actions

    def __str__(self) -> str:
        return (
            self.message or f'Invalid method action: {self.action}\nValid are: {self.valid_actions}'
        )



from mongorepo import exceptions
from mongorepo._base import CollectionProvider
from mongorepo._common import MongorepoDict
from mongorepo.utils import _get_meta_attributes

from ._substitute import _substitute_specific_method
from .methods import SpecificMethod


def _handle_implement_specific_methods[T: type](cls: T, *specific_methods: SpecificMethod) -> T:
    attrs = _get_meta_attributes(cls)
    dto_type = attrs['dto']
    if not dto_type:
        raise exceptions.NoDTOTypeException
    collection = attrs['collection']
    id_field = attrs['id_field']

    if hasattr(cls, '__mongorepo__'):
        __mongorepo__: MongorepoDict = getattr(cls, '__mongorepo__')  # noqa
    else:
        __mongorepo__ = MongorepoDict(
            collection_provider=CollectionProvider(obj=cls, collection=collection), methods={},
        )

    for method in specific_methods:
        implemented_method = _substitute_specific_method(
            cls,
            method=method,
            dto_type=dto_type,
            id_field=id_field,
        )
        __mongorepo__['methods'][method.name] = implemented_method  # type: ignore

        setattr(cls, method.name, __mongorepo__['methods'][method.name])

    setattr(cls, '__mongorepo__', __mongorepo__)
    return cls



import asyncio
from typing import Iterable

from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCollection,
)
from pymongo.client_session import ClientSession
from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo._base import Access, CollectionProvider
from mongorepo._common import MongorepoDict
from mongorepo._methods.impl import (
    AddBatchMethod,
    AddMethod,
    AppendListMethod,
    DeleteMethod,
    GetAllMethod,
    GetListMethod,
    GetListValuesMethod,
    GetMethod,
    IncrementIntegerFieldMethod,
    PopListMethod,
    RemoveListMethod,
    UpdateMethod,
)
from mongorepo._methods.impl_async import (
    AddBatchMethodAsync,
    AddMethodAsync,
    AppendListMethodAsync,
    DeleteMethodAsync,
    GetAllMethodAsync,
    GetListMethodAsync,
    GetListValuesMethodAsync,
    GetMethodAsync,
    IncrementIntegerFieldMethodAsync,
    PopListMethodAsync,
    RemoveListMethodAsync,
    UpdateMethodAsync,
)
from mongorepo.utils import (
    _check_valid_field_type,
    _create_index,
    _create_index_async,
    _get_converter,
    _get_meta_attributes,
    get_prefix,
    raise_exc,
)


def _handle_mongo_repository(
    cls,
    add: bool,
    get: bool,
    add_batch: bool,
    get_all: bool,
    update: bool,
    delete: bool,
    get_list: bool,
    list_fields: Iterable[str] | None,
    integer_fields: Iterable[str] | None,
    method_access: Access | None,
) -> type:
    attributes = _get_meta_attributes(cls)

    collection = attributes['collection']
    if collection is not None and not isinstance(collection, Collection):
        raise exceptions.MongorepoException(
            f'Invalid collection type "{type(collection)}", expected: {Collection}',
        )
    dto = attributes['dto'] or raise_exc(exceptions.NoDTOTypeException)
    index = attributes['index']
    id_field = attributes['id_field']

    prefix = get_prefix(
        access=attributes['method_access'] if not method_access else method_access, cls=cls,
    )

    if index is not None and collection is not None:
        _create_index(index, collection)

    converter = _get_converter(dto, id_field)

    if hasattr(cls, '__mongorepo__'):
        __mongorepo__: MongorepoDict[ClientSession, Collection] = getattr(cls, '__mongorepo__')
    else:
        __mongorepo__ = MongorepoDict[ClientSession, Collection](
            collection_provider=CollectionProvider(obj=cls, collection=collection), methods={},
        )

    if add:
        key = f'{prefix}add'
        add_method = AddMethod(dto, owner=cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = add_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if add_batch:
        key = f'{prefix}add_batch'
        add_batch_method = AddBatchMethod(dto, cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = add_batch_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get:
        key = f'{prefix}get'
        get_method = GetMethod(dto, owner=cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = get_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get_all:
        key = f'{prefix}get_all'
        get_all_method = GetAllMethod(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = get_all_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get_list:
        key = f'{prefix}get_list'
        get_list_method = GetListMethod(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = get_list_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if delete:
        key = f'{prefix}delete'
        delete_method = DeleteMethod(dto, cls)
        __mongorepo__['methods'][key] = delete_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if update:
        key = f'{prefix}update'
        update_method = UpdateMethod(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = update_method
        setattr(cls, key, __mongorepo__['methods'][key])

    if list_fields:
        for field in list_fields:
            _check_valid_field_type(field, dto, list)

            append_method = AppendListMethod(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__append'] = append_method
            setattr(cls, k, __mongorepo__['methods'][k])

            remove_method = RemoveListMethod(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__remove'] = remove_method
            setattr(cls, k, __mongorepo__['methods'][k])

            pop_method = PopListMethod(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__pop'] = pop_method
            setattr(cls, k, __mongorepo__['methods'][k])

            list_values_method = GetListValuesMethod(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__list'] = list_values_method
            setattr(cls, k, __mongorepo__['methods'][k])

    if integer_fields:
        for field in integer_fields:
            _check_valid_field_type(field, dto, int)

            increment_method = IncrementIntegerFieldMethod(
                dto, cls, field_name=field, weight=1,
            )
            __mongorepo__['methods'][k := f'{prefix}incr__{field}'] = increment_method
            setattr(cls, k, __mongorepo__['methods'][k])

            decrement_method = IncrementIntegerFieldMethod(
                dto, cls, field_name=field, weight=-1,
            )
            __mongorepo__['methods'][k := f'{prefix}decr__{field}'] = decrement_method
            setattr(cls, k, __mongorepo__['methods'][k])

    cls.__mongorepo__ = __mongorepo__

    return cls


def _handle_async_mongo_repository(
    cls,
    add: bool,
    add_batch: bool,
    get: bool,
    get_all: bool,
    get_list: bool,
    update: bool,
    delete: bool,
    integer_fields: Iterable[str] | None,
    list_fields: Iterable[str] | None,
    method_access: Access | None,
) -> type:
    """Calls for functions that set different async methods and attributes to
    the class."""
    attributes = _get_meta_attributes(cls)

    collection = attributes['collection']
    if collection is not None and not isinstance(collection, AsyncIOMotorCollection):
        raise exceptions.MongorepoException(
            f'Invalid collection type "{type(collection)}", expected: {AsyncIOMotorCollection}',
        )
    dto = attributes['dto'] or raise_exc(exceptions.NoDTOTypeException)
    index = attributes['index']
    id_field = attributes['id_field']

    prefix = get_prefix(
        access=attributes['method_access'] if not method_access else method_access, cls=cls,
    )

    if hasattr(cls, '__mongorepo__'):
        __mongorepo__: MongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection] = getattr(cls, '__mongorepo__')  # noqa
    else:
        __mongorepo__ = MongorepoDict[AsyncIOMotorClientSession, AsyncIOMotorCollection](
            collection_provider=CollectionProvider(obj=cls, collection=collection), methods={},
        )

    if index is not None and collection is not None:
        asyncio.create_task(_create_index_async(index, collection))

    converter = _get_converter(dto, id_field)

    if add:
        key = f'{prefix}add'
        add_method = AddMethodAsync(dto, owner=cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = add_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if add_batch:
        key = f'{prefix}add_batch'
        add_batch_method = AddBatchMethodAsync(dto, cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = add_batch_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get:
        key = f'{prefix}get'
        get_method = GetMethodAsync(dto, owner=cls, id_field=id_field, converter=converter)
        __mongorepo__['methods'][key] = get_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get_all:
        key = f'{prefix}get_all'
        get_all_method = GetAllMethodAsync(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = get_all_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if get_list:
        key = f'{prefix}get_list'
        get_list_method = GetListMethodAsync(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = get_list_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if delete:
        key = f'{prefix}delete'
        delete_method = DeleteMethodAsync(dto, cls)
        __mongorepo__['methods'][key] = delete_method
        setattr(cls, key, __mongorepo__['methods'][key])
    if update:
        key = f'{prefix}update'
        update_method = UpdateMethodAsync(dto, cls, converter=converter)
        __mongorepo__['methods'][key] = update_method
        setattr(cls, key, __mongorepo__['methods'][key])

    if list_fields:
        for field in list_fields:
            _check_valid_field_type(field, dto, list)

            append_method = AppendListMethodAsync(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__append'] = append_method
            setattr(cls, k, __mongorepo__['methods'][k])

            remove_method = RemoveListMethodAsync(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__remove'] = remove_method
            setattr(cls, k, __mongorepo__['methods'][k])

            pop_method = PopListMethodAsync(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__pop'] = pop_method
            setattr(cls, k, __mongorepo__['methods'][k])

            list_values_method = GetListValuesMethodAsync(dto_type=dto, owner=cls, field_name=field)
            __mongorepo__['methods'][k := f'{prefix}{field}__list'] = list_values_method
            setattr(cls, k, __mongorepo__['methods'][k])

    if integer_fields:
        for field in integer_fields:
            _check_valid_field_type(field, dto, int)

            increment_method = IncrementIntegerFieldMethodAsync(
                dto, cls, field_name=field, weight=1,
            )
            __mongorepo__['methods'][k := f'{prefix}incr__{field}'] = increment_method
            setattr(cls, k, __mongorepo__['methods'][k])

            decrement_method = IncrementIntegerFieldMethodAsync(
                dto, cls, field_name=field, weight=-1,
            )
            __mongorepo__['methods'][k := f'{prefix}decr__{field}'] = decrement_method
            setattr(cls, k, __mongorepo__['methods'][k])

    cls.__mongorepo__ = __mongorepo__

    return cls



import typing as t
from abc import ABC, abstractmethod
from dataclasses import asdict, make_dataclass

from mongorepo._base import Dataclass

P = t.ParamSpec('P')


class ModifierBefore(t.Generic[P], ABC):
    """Abstract base class for modifying input parameters of a `mongorepo`
    method.

    Implement `modify` to transform arguments before execution.
    If no modifications are needed, return them unchanged.
    Other modifiers will use the returned values as input.

    ### NOTE:
    If modifying methods that accept filters:
    1. Filter names will be mapped to dataclass field names. Example:

        @dataclass
        class User:
            name: str

        alias = FieldAlias('name', 'username')

        # Use 'name' instead of 'username'
        def modify(self, name: str):
            ...

    2. If modifying a specific filter, return it as a dictionary:

        def modify(self, id: str) -> dict[str, str]:
            return {'id': id}

    3. If passing multiple arguments unchanged, return them as a tuple in input order:

        def modify(self, a: int, b: str, c: bool, **kwargs) -> tuple[int, str, bool, dict]:
            return (a, b, c, kwargs)

    """

    @abstractmethod
    def modify(self, *args: P.args, **kwargs: P.kwargs) -> t.Any:
        ...


class ModifierAfter[RT](ABC):
    """Abstract base class for modifying the result of a `mongorepo` method.

    Implement `modify` to transform the result before returning it.
    If no modifications are needed, return the input unchanged.
    Other modifiers will use the returned value.

    """

    @abstractmethod
    def modify(self, data: RT) -> t.Any:
        ...


class UpdateSkipModifier[T: Dataclass](ModifierBefore[T]):
    """### Modifier to skip updating dataclass fields values that equal to
    specific value.

    Apply to methods that implement `IUpdateMethod`, `IUpdateMethodAsync` protocols
    * Works with `mongorepo.impelement.methods.UpdateMethod` class

    """

    def __init__(self, skip_if_value: t.Any) -> None:
        self.skip_if_value = skip_if_value

    def modify(self, dto: T, **filters: t.Any) -> tuple[Dataclass, dict[str, t.Any]]:
        _d = asdict(dto)
        new_dataclass = make_dataclass(
            f'{self.__class__.__name__}Dataclass',
            [
                (k, 'typing.Any', f)
                for k, f in dto.__dataclass_fields__.items()
                if _d[k] != self.skip_if_value
            ],
            bases=dto.__class__.__bases__,
        )
        new_dto = new_dataclass(**{k: v for k, v in _d.items() if v != self.skip_if_value})
        return new_dto, filters


class RaiseExceptionModifier(ModifierAfter[t.Any]):
    """Raise exception when result of method execution equals to specific
    value."""
    def __init__(self, exc: Exception | type[Exception], raise_when_result: t.Any) -> None:
        self.exc = exc
        self.raise_when_result = raise_when_result

    def modify(self, data: t.Any) -> t.Any:
        if data == self.raise_when_result:
            raise self.exc
        return data



from typing import Callable

from mongorepo import exceptions
from mongorepo._base import Dataclass
from mongorepo._methods.impl import AddBatchMethod as CallableAddBatchMethod
from mongorepo._methods.impl import AddMethod as CallableAddMethod
from mongorepo._methods.impl import \
    AppendListMethod as CallableAppendListMethod
from mongorepo._methods.impl import DeleteMethod as CallableDeleteMethod
from mongorepo._methods.impl import GetAllMethod as CallableGetAllMethod
from mongorepo._methods.impl import GetListMethod as CallableGetListMethod
from mongorepo._methods.impl import \
    GetListValuesMethod as CallableGetListValuesMethod
from mongorepo._methods.impl import GetMethod as CallableGetMethod
from mongorepo._methods.impl import \
    IncrementIntegerFieldMethod as CallableIncrementIntegerFieldMethod
from mongorepo._methods.impl import PopListMethod as CallablePopListMethod
from mongorepo._methods.impl import \
    RemoveListMethod as CallableRemoveListMethod
from mongorepo._methods.impl import UpdateMethod as CallableUpdateMethod
from mongorepo._methods.impl_async import \
    AddBatchMethodAsync as CallableAddBatchMethodAsync
from mongorepo._methods.impl_async import \
    AddMethodAsync as CallableAddMethodAsync
from mongorepo._methods.impl_async import \
    AppendListMethodAsync as CallableAppendListMethodAsync
from mongorepo._methods.impl_async import \
    DeleteMethodAsync as CallableDeleteMethodAsync
from mongorepo._methods.impl_async import \
    GetAllMethodAsync as CallableGetAllMethodAsync
from mongorepo._methods.impl_async import \
    GetListMethodAsync as CallableGetListMethodAsync
from mongorepo._methods.impl_async import \
    GetListValuesMethodAsync as CallableGetListValuesMethodAsync
from mongorepo._methods.impl_async import \
    GetMethodAsync as CallableGetMethodAsync
from mongorepo._methods.impl_async import \
    IncrementIntegerFieldMethodAsync as \
    CallableIncrementIntegerFieldMethodAsync
from mongorepo._methods.impl_async import \
    PopListMethodAsync as CallablePopListMethodAsync
from mongorepo._methods.impl_async import \
    RemoveListMethodAsync as CallableRemoveListMethodAsync
from mongorepo._methods.impl_async import \
    UpdateMethodAsync as CallableUpdateMethodAsync
from mongorepo.implement.methods import (
    AddBatchMethod,
    AddMethod,
    DeleteMethod,
    GetAllMethod,
    GetListMethod,
    GetMethod,
    IncrementIntegerFieldMethod,
    ListAppendMethod,
    ListGetFieldValuesMethod,
    ListPopMethod,
    ListRemoveMethod,
    SpecificMethod,
    UpdateMethod,
)


def implement_mapper(specific_method: SpecificMethod) -> type:
    """Map implement method to mongorepo implementation."""
    s = specific_method
    match specific_method.__class__.__name__:
        case GetMethod.__name__:
            return CallableGetMethodAsync if s.is_async else CallableGetMethod
        case GetAllMethod.__name__:
            return CallableGetAllMethodAsync if s.is_async else CallableGetAllMethod
        case GetListMethod.__name__:
            return CallableGetListMethodAsync if s.is_async else CallableGetListMethod
        case AddBatchMethod.__name__:
            return CallableAddBatchMethodAsync if s.is_async else CallableAddBatchMethod
        case AddMethod.__name__:
            return CallableAddMethodAsync if s.is_async else CallableAddMethod
        case DeleteMethod.__name__:
            return CallableDeleteMethodAsync if s.is_async else CallableDeleteMethod
        case UpdateMethod.__name__:
            return CallableUpdateMethodAsync if s.is_async else CallableUpdateMethod

        case ListAppendMethod.__name__:
            return CallableAppendListMethodAsync if s.is_async else CallableAppendListMethod
        case ListRemoveMethod.__name__:
            return CallableRemoveListMethodAsync if s.is_async else CallableRemoveListMethod
        case ListPopMethod.__name__:
            return CallablePopListMethodAsync if s.is_async else CallablePopListMethod
        case ListGetFieldValuesMethod.__name__:
            return CallableGetListValuesMethodAsync if s.is_async else CallableGetListValuesMethod

        case IncrementIntegerFieldMethod.__name__:
            if s.is_async:
                return CallableIncrementIntegerFieldMethodAsync
            else:
                return CallableIncrementIntegerFieldMethod

    raise exceptions.MongorepoException(
        f'Cannot map specific method {specific_method.__class__} to mongorepo implementation',
    )


def initialize_callable_mongorepo_method(
    mongorepo_method: type,
    owner: type,
    dto_type: type[Dataclass],
    field_name: str | None = None,
    integer_weight: int | None = None,
    **kwargs,
) -> Callable:
    match (mcls := mongorepo_method.__name__):
        case CallableAddMethod.__name__ | CallableAddMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)
        case CallableGetMethod.__name__ | CallableGetMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)
        case CallableGetListMethod.__name__ | CallableGetListMethodAsync.__name__:
            return mongorepo_method(dto_type, owner=owner, **kwargs)
        case CallableAddBatchMethod.__name__ | CallableAddBatchMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)
        case CallableGetAllMethod.__name__ | CallableGetAllMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)
        case CallableUpdateMethod.__name__ | CallableUpdateMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)
        case CallableDeleteMethod.__name__ | CallableDeleteMethodAsync.__name__:
            return mongorepo_method(dto_type=dto_type, owner=owner, **kwargs)

        case CallablePopListMethod.__name__ | CallablePopListMethodAsync.__name__:
            if not field_name:
                raise exceptions.MongorepoException(
                    f'Cannot initialize {mcls}: field_name was not provided',
                )
            return mongorepo_method(dto_type=dto_type, owner=owner, field_name=field_name, **kwargs)
        case CallableAppendListMethod.__name__ | CallableAppendListMethodAsync.__name__:
            if not field_name:
                raise exceptions.MongorepoException(
                    f'Cannot initialize {mcls}: field_name was not provided',
                )
            return mongorepo_method(dto_type=dto_type, owner=owner, field_name=field_name, **kwargs)
        case CallableRemoveListMethod.__name__ | CallableRemoveListMethodAsync.__name__:
            if not field_name:
                raise exceptions.MongorepoException(
                    f'Cannot initialize {mcls}: field_name was not provided',
                )
            return mongorepo_method(dto_type=dto_type, owner=owner, field_name=field_name, **kwargs)
        case CallableGetListValuesMethod.__name__ | CallableGetListValuesMethodAsync.__name__:
            if not field_name:
                raise exceptions.MongorepoException(
                    f'Cannot initialize {mcls}: field_name was not provided',
                )
            return mongorepo_method(dto_type=dto_type, owner=owner, field_name=field_name, **kwargs)

        case (
            CallableIncrementIntegerFieldMethod.__name__
            | CallableIncrementIntegerFieldMethodAsync.__name__
        ):
            if not field_name or not integer_weight:
                raise exceptions.MongorepoException(
                    f'Cannot initialize {mcls}: field_name or weight was not provided: '
                    f'{field_name=}, {integer_weight=}',
                )
            return mongorepo_method(dto_type, owner, field_name, weight=integer_weight, **kwargs)

    raise exceptions.MongorepoException(
        f'Cannot initialize {mcls}: {mcls} is not implemented',
    )



import typing as t

from mongorepo._base import Dataclass, SessionType

if t.TYPE_CHECKING:
    from pymongo.results import InsertManyResult, UpdateResult


class MongorepoMethod(t.Protocol[SessionType]):
    session: SessionType | None


class IGetMethod[T: Dataclass](t.Protocol):
    def __call__(self, **filters: t.Any) -> T | None:
        ...


class IGetMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, **filters: t.Any) -> T | None:
        ...


class IAddMethod[T: Dataclass](t.Protocol):
    def __call__(self, dto: T) -> T:
        ...


class IAddMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, dto: T) -> T:
        ...


class IAddBatchMethod[T: Dataclass](t.Protocol):
    def __call__(self, dto_list: list[T]) -> 'InsertManyResult':
        ...


class IAddBatchMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, dto_list: list[T]) -> 'InsertManyResult':
        ...


class IGetAllMethod[T: Dataclass](t.Protocol):
    def __call__(self, **filters: t.Any) -> t.Generator[T, None, None]:
        ...


class IGetAllMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, **filters: t.Any) -> t.AsyncGenerator[T, None]:
        ...


class IGetListMethod[T: Dataclass](t.Protocol):
    def __call__(self, offset: int, limit: int, **filters: t.Any) -> list[T]:
        ...


class IGetListMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, offset: int, limit: int, **filters: t.Any) -> list[T]:
        ...


class IDeleteMethod(t.Protocol):
    def __call__(self, **filters: t.Any) -> bool:
        ...


class IDeleteMethodAsync(t.Protocol):
    async def __call__(self, **filters: t.Any) -> bool:
        ...


class IUpdateMethod[T: Dataclass](t.Protocol):
    def __call__(self, dto: T, **filters: t.Any) -> T | None:
        ...


class IUpdateMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, dto: T, **filters: t.Any) -> T | None:
        ...


class IUpdateArrayMethod[T: Dataclass](t.Protocol):
    def __call__(self, value: t.Any, **filters: t.Any) -> 'UpdateResult':
        ...


class IUpdateArrayMethodAsync[T: Dataclass](t.Protocol):
    async def __call__(self, value: t.Any, **filters: t.Any) -> 'UpdateResult':
        ...


class IPopListMethod[T](t.Protocol):
    def __call__(self, **filters: t.Any) -> T | None:
        ...


class IPopListMethodAsync[T](t.Protocol):
    async def __call__(self, **filters: t.Any) -> T | None:
        ...


class IGetListValuesMethod[T](t.Protocol):
    def __call__(
        self, offset: int = 0, limit: int = 20, **filters: t.Any,
    ) -> list[T] | list[t.Any] | None:
        ...


class IGetListValuesMethodAsync[T](t.Protocol):
    async def __call__(
        self, offset: int = 0, limit: int = 20, **filters: t.Any,
    ) -> list[T] | list[t.Any] | None:
        ...


class IIncrementIntegerFieldMethod(t.Protocol):
    def __call__(self, weight: int | None = None, **filters) -> 'UpdateResult':
        ...


class IIncrementIntegerFieldMethodAsync(t.Protocol):
    async def __call__(self, weight: int | None = None, **filters) -> 'UpdateResult':
        ...



from typing import Literal

Operation = Literal[
    '$set',          # Set value
    '$push',         # Append to an array
    '$incr',         # Increment value
    '$pull',         # Remove an element from an array
    '$each',         # Apply operation to each element
    '$sort',         # Sort elements
    '$slice',        # Slice elements
    '$addToSet',     # Add unique values to an array
    '$pop',          # Remove first or last element
    '$rename',       # Rename a field
    '$unset',        # Remove a field
    '$currentDate',  # Set the field to the current date
    '$min',          # Update to a minimum value if specified value is less
    '$max',          # Update to a maximum value if specified value is more
    '$mul',          # Multiply a field value by a number
]

Condition = Literal[
    '$and',          # Logical AND
    '$or',           # Logical OR
    '$not',          # Logical NOT
    '$nor',          # Logical NOR
    '$exists',       # Field existence check
    '$type',         # Type check for field
    '$in',           # Field value is in array
    '$nin',          # Field value is not in array
    '$eq',           # Equals
    '$ne',           # Not equals
    '$gt',           # Greater than
    '$gte',          # Greater than or equal
    '$lt',           # Less than
    '$lte',          # Less than or equal
    '$regex',        # Regular expression match
    '$all',          # Array contains all elements
    '$elemMatch',    # Match elements in an array field
    '$size',         # Match array size
]


AggregationStage = Literal[
    '$match',          # Filters documents
    '$group',          # Groups documents by a specified expression
    '$sort',           # Sorts documents by a specified field
    '$project',        # Selects which fields to include or exclude
    '$limit',          # Limits the number of documents
    '$skip',           # Skips a specified number of documents
    '$unwind',         # Deconstructs an array field to output one document per element
    '$lookup',         # Performs a join with another collection
    '$facet',          # Processes multiple pipelines within a single stage
    '$count',          # Counts the number of documents
    '$addFields',      # Adds new fields to documents
    '$replaceRoot',    # Replaces the root document with a specified sub-document
    '$bucket',         # Categorizes documents into buckets
    '$bucketAuto',     # Categorizes documents into auto-generated buckets
    '$geoNear',        # Sorts documents by proximity to a specified point
]

UpdateModifier = Literal[
    '$inc',            # Increments a field by a specified value
    '$mul',            # Multiplies the value of a field by a specified number
    '$rename',         # Renames a field
    '$setOnInsert',    # Sets a field if document is inserted
    '$set',            # Sets the value of a field
    '$unset',          # Removes a field from a document
    '$min',            # Only updates the field if the specified value is less
    '$max',            # Only updates the field if the specified value is greater
    '$currentDate',    # Sets the value of a field to the current date
    '$addToSet',       # Adds value to an array if it doesn’t exist
    '$push',           # Appends a value to an array
    '$pull',           # Removes elements from an array that match a condition
    '$pullAll',        # Removes all instances of specified values from an array
    '$pop',            # Removes the first or last item of an array
    '$bit',            # Performs bitwise operations on integer values
]



from enum import StrEnum
from typing import Literal


class MethodAction(StrEnum):
    GET = 'get'
    GET_LIST = 'get_list'
    GET_ALL = 'get_all'
    UPDATE = 'update'
    ADD = 'add'
    ADD_BATCH = 'add_batch'
    DELETE = 'delete'

    INTEGER_INCREMENT = 'incr__'
    INTEGER_DECREMENT = 'decr__'

    LIST_APPEND = '__append'
    LIST_REMOVE = '__remove'
    LIST_POP = '__pop'
    LIST_FIELD_VALUES = '__list'

    def as_mongo_action(self) -> str:
        match self:
            case self.LIST_APPEND:
                return '$push'
            case self.LIST_REMOVE:
                return '$pull'
            case self.LIST_POP:
                return '$pop'
        return self.value


class ParameterEnum(StrEnum):
    FILTER = 'filters'
    OFFSET = 'offset'
    LIMIT = 'limit'
    DTO = 'dto'
    VALUE = 'value'
    WEIGHT = 'weight'
    FILTER_ALIAS = '__filter_alias'


LParameter = Literal[
    ParameterEnum.FILTER,
    ParameterEnum.OFFSET,
    ParameterEnum.LIMIT,
    ParameterEnum.DTO,
    ParameterEnum.VALUE,
    ParameterEnum.WEIGHT,
    ParameterEnum.FILTER_ALIAS,
]
